/*-- 

 Copyright (C) @year@ Brett McLaughlin & Jason Hunter. All rights reserved.
 
 Redistribution and use in source and binary forms, with or without modifica-
 tion, are permitted provided that the following conditions are met:
 
 1. Redistributions of source code must retain the above copyright notice,
    this list of conditions, and the following disclaimer.
 
 2. Redistributions in binary form must reproduce the above copyright notice,
    this list of conditions, the disclaimer that follows these conditions,
    and/or other materials provided with the distribution.
 
 3. The names "JDOM" and "Java Document Object Model" must not be used to
    endorse or promote products derived from this software without prior
    written permission. For written permission, please contact
    license@jdom.org.
 
 4. Products derived from this software may not be called "JDOM", nor may
    "JDOM" appear in their name, without prior written permission from the
    JDOM Project Management (pm@jdom.org).
 
 THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED WARRANTIES,
 INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 FITNESS  FOR A PARTICULAR  PURPOSE ARE  DISCLAIMED.  IN NO  EVENT SHALL  THE
 JDOM PROJECT  OR ITS CONTRIBUTORS  BE LIABLE FOR  ANY DIRECT, INDIRECT, 
 INCIDENTAL, SPECIAL,  EXEMPLARY, OR CONSEQUENTIAL  DAMAGES (INCLUDING, BUT 
 NOT LIMITED TO, PROCUREMENT  OF SUBSTITUTE GOODS OR SERVICES; LOSS
 OF USE, DATA, OR  PROFITS; OR BUSINESS  INTERRUPTION)  HOWEVER CAUSED AND ON
 ANY  THEORY OF LIABILITY,  WHETHER  IN CONTRACT,  STRICT LIABILITY,  OR TORT
 (INCLUDING  NEGLIGENCE OR  OTHERWISE) ARISING IN  ANY WAY OUT OF THE  USE OF
 THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 
 This software  consists of voluntary contributions made  by many individuals
 on  behalf of the Java Document Object Model Project and was originally 
 created by Brett McLaughlin <brett@jdom.org> and 
 Jason Hunter <jhunter@jdom.org>. For more  information on the JDOM 
 Project, please see <http://www.jdom.org/>.
 
 */
package org.jdom;

import java.util.LinkedList;
import java.util.Iterator;
import java.util.List;

/**
 * <p><code>Attribute</code> defines behavior for an XML 
 *   attribute, modeled in Java.  Methods allow the user 
 *   to obtain the value of the element's textual content,
 *   obtain its attributes, and get its children.
 * </p> 
 *
 *
 * @author Brett McLaughlin
 * @author Jason Hunter
 * @version 1.0 
 */
public class Element {

    /** The local name of the <code>Element</code> */
    protected String name;
    
    /** The namespace prefix of the <code>Element</code> */
    protected String prefix;
    
    /** The attributes of the <code>Element</code> */
    protected List attributes;    

    /** The mixed content of the <code>Element</code> */
    protected List content;
    
    /*
     * XXX: We need to build in a "cache" with the child elements -
     *  They are accessed so frequently we could improve out
     *  performance big time with this. (brett)
     */
    
    /**
     * <p>
     * Default, no-args constructor for implementations
     *   to use if needed.
     * </p>
     */
    protected Element() { }
    
    /**
     * <p>
     * This will create a new <code>Element</code>
     *   with the supplied namespace prefix and (local)
     *   name.
     * </p>
     *
     * @param prefix <code>String</code> prefix for element.
     * @param name <code>String</code> name of element.     
     */
    public Element(String prefix, String name) {
        this.name = name;
        this.prefix = prefix;
        attributes = new LinkedList();
        content = new LinkedList();
    }
    
    /**
     * <p>
     * This will create a new <code>Element</code>
     *   with the supplied (local) name.  This assumes
     *   that no prefix should be issued for the
     *   <code>Element</code>, and that it is in the
     *   default namespace.
     * </p>
     *
     * @param name <code>String</code> name of element.
     */
    public Element(String name) {
        int split = name.indexOf(":");
        if (split < 0) {
            this.name = name;
            this.prefix = "";
        } else {
            if (!name.startsWith(":")) {
                this.prefix = name.substring(0, split);
            } else {
                prefix = "";
            }
            this.name = name.substring(split+1);
        }
        attributes = new LinkedList();
        content = new LinkedList();
    }
    
    /**
     * <p>
     * This returns the (local) name of the
     *   <code>Element</code>, without any
     *   namespace prefix, if one exists.
     * </p>
     * 
     * @return <code>String</code> - element name.
     */
    public String getName() {
        return name;
    }
    
    /**
     * <p>
     * This returns the namespace prefix 
     *   of the <code>Element</code>, if 
     *   one exists.  Otherwise, an empty
     *   <code>String</code> is returned.
     * </p>
     * 
     * @return <code>String</code> - namespace prefix.
     */    
    public String getNamespacePrefix() {
        return prefix;
    }

    /**
     * <p>
     * This returns the full name of the
     *   <code>Element</code>, in the form
     *   [namespacePrefix]:[localName].  If
     *   no namespace prefix exists for the
     *   <code>Element</code>, simply the 
     *   local name is returned.
     * </p>
     *
     * @return <code>String</code> - full name of element.
     */
    public String getFullName() {
        if (prefix.equals("")) {
            return getName();
        }
        
        return new StringBuffer(prefix)
            .append(":")
            .append(name)
            .toString();
    }    
    
    /**
     * <p>
     * This will set the name of the element.  If the name specified
     *   has a namespace prefix (indicated by a colon), then the
     *   prefix is split from the element and handled.  If no namespace
     *   prefix is supplied, the namespace prefix is empty.
     * </p>
     *
     * @param name <code>String</code> full name of element.
     * @return <code>Element</code> - this element modified.
     */
    public Element setName(String name) {
        int split = name.indexOf(":");
        if (split < 0) {
            prefix = "";
            this.name = name;            
        } else {
            if (!name.startsWith(":")) {
                prefix = name.substring(0, split);
            } else {
                prefix = "";
            }
            this.name = name.substring(split+1);
        }
        
        return this;
    }
    
    /**
     * <p>
     * This will set the name of the element.  It allows
     *   a namespace prefix to be passed in as well as the local
     *   name of the element.
     * </p>
     *
     * @param prefix <code>String</code> prefix for name.
     * @param name <code>String</code> local name of element.
     * @return <code>Element</code> - this element modified.
     */
    public Element setName(String prefix, String name) {
        this.prefix = prefix;
        this.name = name;
        
        return this;
    }
    
    /**
     * <p>
     * This will return the actual textual content of this 
     *   <code>Element</code>.  This will include all text within
     *   this single element, including <code>CDATA</code> sections
     *   if they exist.  If no textual value exists for the 
     *   <code>Element</code>, an empty <code>String</code> is returned.
     * </p>
     *
     * @return <code>String</code> - value for this element.
     */
    public String getContent() {
        StringBuffer textContent = new StringBuffer();
        
        Iterator i = content.iterator();
        while (i.hasNext()) {
            Object obj = i.next();
            if (obj instanceof String) {
                textContent.append((String)obj);
            }
        }
        
        return textContent.toString();
    }   
    
    /**
     * <p>
     * This will set the textual content of the <code>Element</code>.
     *   If this <code>Element</code> will have both textual content
     *   and nexted elements, <code>{@link #setMixedContent}</code>
     *   should be used instead,
     * </p>
     *
     * @param textContent <code>String</code> content for <code>Element</code>.
     * @return <code>Element</code> - this element modified.
     */
    public Element setContent(String textContent) {
        content.clear();
        content.add(textContent);        
        
        return this;
    }
    
    /**
     * <p>
     * This will indicate whether the element has mixed content or not.
     *   Mixed content is when an element contains both textual and
     *   element data within it.  When this evaluates to <code>true</code>,
     *   <code>{@link #getMixedContent}</code> should be used for getting
     *   element data.
     * </p>
     * <p>
     * <em>Note:</em>I think there's a better way to do this, 
     *   maybe store a member variable, but I'm too tired
     *   to really think of what it is... (brett).
     * </p>     
     *
     * @return <code>boolean</code> - indicating whether there
     *         is mixed content (both textual data and elements).
     */
    public boolean hasMixedContent() {
        boolean hasText = false;
        boolean hasElements = false;
        boolean hasComments = false;
        boolean hasEntities = false;        
        
        Iterator i = content.iterator();
        while (i.hasNext()) {
            Object obj = i.next();
            if (obj instanceof String) {
                if (hasElements || 
                    hasComments || 
                    hasEntities) {
                    return true;
                }
                hasText = true;
            } else if (obj instanceof Element) {
                if (hasText || 
                    hasComments || 
                    hasEntities) {
                    return true;
                }
                hasElements = true;                
            } else if (obj instanceof Comment) {
                if (hasText || 
                    hasElements || 
                    hasEntities) {
                    return true;
                }
                hasComments = true;                
            } else if (obj instanceof Entity) {
                if (hasText || 
                    hasElements || 
                    hasComments) {
                    return true;
                }
                hasEntities = true;                
            }       
        }
        
        return false;
    }    
    
    /**
     * <p>
     * This will return the content of the element.  This should be
     *   used when the <code>{@link #hasMixedContent}</code>
     *   evaluates to <code>true</code>.  When there is no mixed
     *   content, it returns a <code>List</code> with a single
     *   <code>String</code> (when only data is present) or a 
     *   <code>List</code> with only elements (when only nested
     *   elements are present).
     * </p>
     *
     * @return <code>List</code> - the mixed content of the 
     *         <code>Element</code>: contains <code>String</code>,
     *         <code>Element</code>, <code>{@link Comment}</code>,
     *         and <code>{@link Entity}</code> objects.
     */
    public List getMixedContent() {
        return new PartialList(content);
    }
    
    /**
     * <p>
     * This will return the content of the element.  This should be
     *   used when the <code>{@link #hasMixedContent}</code>
     *   evaluates to <code>true</code>.  When there is no mixed
     *   content, it returns a <code>List</code> with a single
     *   <code>String</code> (when only data is present) or a 
     *   <code>List</code> with only elements (when only nested
     *   elements are present).
     * </p>
     *
     * @return <code>Element</code> - this element modified.
     */
    public Element setMixedContent(List mixedContent) {
        content.clear();
        content.addAll(mixedContent);
        
        return this;
    }
    
    /**
     * <p>
     * This will return a <code>List</code> of all the XML 
     *   elements nested directly (one level deep) within 
     *   this <code>Element</code>, each in <code>Element</code> 
     *   form.  If the <code>Element</code> has no nested elements, 
     *   an empty list will be returned.
     * </p><p>
     * This performs no recursion, so an elements nested two levels
     *   deep would have to be obtained with:
     * <pre>
     * <code>
     *   List nestedElements = currentElement.getChildren();
     *   for (int i=0; i&lt;nestedElements.size(); i++) {
     *     Element oneLevelDeep = (Element)nestedElements.get(i);
     *     List twoLevelsDeep = oneLevelDeep.getChildren();
     *     // Do something with these children
     *   }
     * </code>
     * </pre>
     * </p>
     *
     * @return <code>List</code> - list of nested 
     *                               <code>Element</code>
     *                               instances for this element.
     */
    public List getChildren() {
        PartialList elements = new PartialList(content);

        Iterator i = content.iterator();
        while (i.hasNext()) {
            Object obj = i.next();
            if (obj instanceof Element) {
                elements.addPartial(obj);
            }
        }
        
        return elements;
    }
    
    /**
     * <p>
     * This will set the children of this <code>Element</code> to the
     *   <code>Element</code>s within the supplied <code>:ost</code>.
     *   All existing children of this <code>Element</code> are replaced.
     * </p>
     *
     * @param children <code>List</code> of <code>Element</code>s to add.
     * @return <code>Element</code> - this element modified.
     */
    public Element setChildren(List children) {
        return setMixedContent(children);
    }  
    
    /**
     * <p>
     * This will return a <code>List</code> of all the XML 
     *   elements nested directly (one level deep) within 
     *   this <code>Element</code> whose names match the name
     *   specified, each in <code>Element</code> 
     *   form.  If the <code>Element</code> has no nested elements
     *   that match the requested name, an empty list will be returned.
     * </p><p>
     * Either the raw name (XML 1.0: namespace prefix and local name) of
     *   the element can be provided (for example, JavaXML:Book), or simply
     *   the local name (XML Namespace aware: local name only) of the 
     *   element (for example, Book), can be provided.
     * </p><p>
     * Please see the notes for <code>{@link #getChildren}</code>
     *   regarding how deep this search will go within element
     *   nestings. Only elements directly nested within this <code>Element</code>
     *   will be returned.
     * </p>
     *
     *
     * @param name <code>String</code> name of attribute to return.
     * @return <code>List</code> - all matching child elements.
     */
    public List getChildren(String name) {          
        PartialList children = new PartialList(getChildren());
        
        // If the name is a fullName (with prefix, handle it
        boolean hasNamespace = (name.indexOf(":") > 0);         
        
        Iterator i = content.iterator();
        while (i.hasNext()) {
            Object obj = i.next();
            if (obj instanceof Element) {
                Element element = (Element)obj;
                if (hasNamespace) {
                    if (element.getFullName().equals(name)) {
                        children.addPartial(element);
                    }
                } else {
                    if (element.getName().equals(name)) {
                        children.addPartial(element);
                    }
                }
            }
        }
        
        return children;
    }
    
    /**
     * <p>
     * This will return the children <code>{@link Element}</code>s for the
     *   specified element name on this <code>Element</code>.  If
     *   multiple elements exist for the specified name, all are
     *   returned.
     * </p><p>
     * Please see the notes for <code>{@link #getChildren}</code>
     *   regarding how deep this search will go within element
     *   nestings. Only elements directly nested within this <code>Element</code>
     *   will be returned.
     * </p>
     *
     * @param prefix <code>String</code> prefix of elements to find
     * @param name <code>String</code> name of child elements to return.
     * @return <code>List</code> - the requested elements.
     */
    public List getChildren(String prefix, String name) {
        PartialList children = new PartialList(getChildren());       
        
        Iterator i = content.iterator();
        while (i.hasNext()) {
            Object obj = i.next();
            if (obj instanceof Element) {
                Element element = (Element)obj;
                if ((element.getNamespacePrefix().equals(prefix)) &&
                    (element.getName().equals(name))) {
                    children.addPartial(element);
                }
            }
        }
        
        return children;        
    }
    
    /**
     * <p>
     * This will return a <code>List</code> of all the XML 
     *   elements nested directly (one level deep) within 
     *   this <code>Element</code> whose names match the name
     *   specified, each in <code>Element</code> 
     *   form.  If the <code>Element</code> has no nested elements
     *   that match the requested name, an empty list will be returned.
     * </p><p>
     * Either the raw name (XML 1.0: namespace prefix and local name) of
     *   the element can be provided (for example, JavaXML:Book), or simply
     *   the local name (XML Namespace aware: local name only) of the 
     *   element (for example, Book), can be provided.
     * </p><p>
     * Please see the notes for <code>{@link #getChildren}</code>
     *   regarding how deep this search will go within element
     *   nestings. Only elements directly nested within this <code>Element</code>
     *   will be returned.
     * </p>
     *
     * @param name <code>String</code> name of attribute to return.     
     * @return <code>Element</code> - the first matching child element.
     * @throws <code>NoSuchElementException</code> - when the requested
     *         XML element does not exist within this <code>Element</code>.     
     */
    public Element getChild(String name) throws NoSuchElementException {          
        // If the name is a fullName (with prefix, handle it
        boolean hasNamespace = (name.indexOf(":") > 0);         
        
        Iterator i = content.iterator();
        while (i.hasNext()) {
            Object obj = i.next();
            if (obj instanceof Element) {
                Element element = (Element)obj;
                if (hasNamespace) {
                    if (element.getFullName().equals(name)) {
                        return element;
                    }
                } else {
                    if (element.getName().equals(name)) {
                        return element;
                    }
                }
            }
        }
        
        // If we got here, none found
        throw new NoSuchElementException(name);
    }    
    
    /**
     * <p>
     * This will return the child <code>{@link Element}</code>s for the
     *   specified element name on this <code>Element</code>.  If
     *   multiple elements exist for the specified name, only the first is
     *   returned.
     * </p><p>
     * Please see the notes for <code>{@link #getChildren}</code>
     *   regarding how deep this search will go within element
     *   nestings. Only elements directly nested within this <code>Element</code>
     *   will be returned.
     * </p>
     *
     * @param prefix <code>String</code> prefix of elements to find
     * @param name <code>String</code> name of child elements to return.
     * @return <code>Element</code> - the first matching child element.
     * @throws <code>NoSuchElementException</code> - when the requested
     *         XML element does not exist within this <code>Element</code>.     
     */
    public Element getChild(String prefix, String name) 
        throws NoSuchElementException {
            
        Iterator i = content.iterator();
        while (i.hasNext()) {
            Object obj = i.next();
            if (obj instanceof Element) {
                Element element = (Element)obj;
                if ((element.getNamespacePrefix().equals(prefix)) &&
                    (element.getName().equals(name))) {
                    return element;
                }
            }
        }            
    
        // If we got here, none found
        throw new NoSuchElementException(prefix + ":" + name);
    }
    
    /**
     * <p>
     * This will add test to the content of this
     *   <code>Element</code>.
     * </p>
     *
     * @param text <code>String</code> to add as content.
     * @return <code>Element</code> - this element modified.
     */
    public Element addText(String text) {
        content.add(text);
        
        return this;
    }    
    
    /**
     * <p>
     * This will add an <code>Element</code> as a child of this
     *   <code>Element</code>.
     * </p>
     *
     * @param element <code>Element</code> to add as a child.
     * @return <code>Element</code> - this element modified.
     */
    public Element addChild(Element element) {
        content.add(element);
        
        return this;
    }
    
    /**
     * <p>
     * This will add an <code>Entity</code> as a child of this
     *   <code>Element</code>.
     * </p>
     *
     * @param entity <code>Entity</code> to add as a child.
     * @return <code>Element</code> - this element modified.
     */
    public Element addChild(Entity entity) {
        content.add(entity);
        
        return this;
    }    
    
    /**
     * <p>
     * This will add a <code>String</code> as a child of this
     *   <code>Element</code>.
     * </p>
     *
     * @param s <code>String</code> to add as a child.
     * @return <code>Element</code> - this element modified.
     */
    public Element addChild(String s) {
        content.add(s);
        
        return this;
    }
    
    /**
     * <p>
     * This will remove the specified <code>Element</code>.
     * </p>
     *
     * @param child <code>Element</code> to delete.
     * @return <code>boolean</code> - whether deletion occurred.
     */
    public boolean removeChild(Element element) {
        return content.remove(element);
    }
    
    /**
     * <p>
     * This will remove the specified <code>Entity</code>.
     * </p>
     *
     * @param child <code>Entity</code> to delete.
     * @return <code>boolean</code> - whether deletion occurred.
     */
    public boolean removeChild(Entity entity) {
        return content.remove(entity);
    }    
    
    /**
     * <p>
     * This will remove the <code>Element</code> with the
     *   specified name.  If multiple <code>Element</code>s 
     *   exist with that name, the first matching child is
     *   removed.
     * </p>
     * <p>
     * <em>Note:</em>Currently, this does not assume that you want the
     *   default namespace.  I can perform 
     *   <code>removeChild("foo:bar")</code>, and it does the right
     *   thing.  But if I have a document with elements 
     *   <code>&lt;foo:bar&gt;</code> as well as <code>&lt;bar&gt;</code>,
     *   this currently removes the <i>first</i> of those when
     *   <code>removeChild("bar")</code> is invoked.  Do we want to
     *   explicitly enforce the default namespace (none) if no prefix
     *   is given? (brett)     
     * </p>     
     *
     * @param name <code>String</code> name of child to delete.
     * @return <code>boolean</code> - whether deletion occurred.
     */
    public boolean removeChild(String name) {
        // If the name is a fullName (with prefix, handle it
        boolean hasNamespace = (name.indexOf(":") > 0); 
        
        Iterator i = content.iterator();
        while (i.hasNext()) {
            Object obj = i.next();
            if (obj instanceof Element) {
                Element element = (Element)obj;
                if (hasNamespace) {
                    if (element.getFullName().equals(name)) {
                        i.remove();
                        return true;
                    }
                } else {
                    if (element.getName().equals(name)) {
                        i.remove();
                        return true;
                    }
                }
            } 
        }
        
        // If we got here, none found
        return false;
    }
    
    /**
     * <p>
     * This will remove the <code>Element</code> with the
     *   specified name.  If multiple <code>Element</code>s 
     *   exist with that name, the first matching child is
     *   removed.
     * </p>
     *
     * @param prefix <code>String</code> prefix of element to remove.
     * @param name <code>String</code> local name of child to delete.
     * @return <code>boolean</code> - whether deletion occurred.
     */
    public boolean removeChild(String prefix, String name) {    
        Iterator i = content.iterator();
        while (i.hasNext()) {
            Object obj = i.next();
            if (obj instanceof Element) {
                Element element = (Element)obj;
                if ((element.getNamespacePrefix().equals(prefix)) &&
                    (element.getName().equals(name))) {
                    i.remove();
                    return true;
                }
            }
        }
        
        // If we got here, none found
        return false;        
    }
    
    /**
     * <p>
     * This will remove the <code>Element</code> with the
     *   specified name.  If multiple <code>Element</code>s 
     *   exist with that name, all matching children are
     *   removed.
     * </p>
     * <p>
     * <em>Note:</em>Currently, this does not assume that you want the
     *   default namespace.  I can perform 
     *   <code>removeChild("foo:bar")</code>, and it does the right
     *   thing.  But if I have a document with elements 
     *   <code>&lt;foo:bar&gt;</code> as well as <code>&lt;bar&gt;</code>,
     *   this currently removes the <i>first</i> of those when
     *   <code>removeChild("bar")</code> is invoked.  Do we want to
     *   explicitly enforce the default namespace (none) if no prefix
     *   is given? (brett)     
     * </p>     
     *
     * @param name <code>String</code> name of child to delete.
     * @return <code>boolean</code> - whether deletion occurred.
     */
    public boolean removeChildren(String name) {
        boolean deletedSome = false;
        
        // If the name is a fullName (with prefix, handle it
        boolean hasNamespace = (name.indexOf(":") > 0); 
        
        Iterator i = content.iterator();
        while (i.hasNext()) {
            Object obj = i.next();
            if (obj instanceof Element) {
                Element element = (Element)obj;
                if (hasNamespace) {
                    if (element.getFullName().equals(name)) {
                        i.remove();
                        deletedSome = true;
                    }
                } else {
                    if (element.getName().equals(name)) {
                        i.remove();
                        deletedSome = true;
                    }
                }
            }
        }
        
        return deletedSome;
    }
    
    /**
     * <p>
     * This will remove the <code>Element</code> with the
     *   specified name.  If multiple <code>Element</code>s 
     *   exist with that name, the first matching child is
     *   removed.
     * </p>
     *
     * @param prefix <code>String</code> prefix of element to remove.
     * @param name <code>String</code> local name of child to delete.
     * @return <code>boolean</code> - whether deletion occurred.
     */
    public boolean removeChildren(String prefix, String name) {    
        boolean deletedSome = false;
        
        Iterator i = content.iterator();
        while (i.hasNext()) {
            Object obj = i.next();
            if (obj instanceof Element) {
                Element element = (Element)obj;
                if ((element.getNamespacePrefix().equals(prefix)) &&
                    (element.getName().equals(name))) {
                    i.remove();
                    deletedSome = true;
                }
            }
        }
        
        return deletedSome;
    }
    
    /**
     * <p>
     * This removes all children for this <code>Element</code>.
     * </p>
     *
     * @return <code>boolean</code> - whether deletion occurred.
     */
    public boolean removeChildren() {
        Iterator i = content.iterator();
        while (i.hasNext()) {
            Object obj = i.next();
            if (obj instanceof Element) {
                i.remove();
            }
        }
        
        return true;
    }
    
    /**
     * <p>
     * This will add a <code>{@link Comment}</code> as a child of this
     *   <code>Element</code>.
     * </p>
     *
     * @param comment <code>Comment</code> to add as a child.
     * @return <code>Element</code> - this element modified.
     */
    public Element addChild(Comment comment) {
        content.add(comment);
        
        return this;
    }
    
    /**
     * <p>
     * This will remove the specified <code>Comment</code>.
     * </p>     
     *
     * @param comment <code>Comment</code> to delete.
     * @return <code>boolean</code> - whether deletion occurred.
     */
    public boolean removeChild(Comment comment) {
        return content.remove(comment);
    }        
    
    /**
     * <p>
     * This returns the complete set of 
     *   <code>{@link Attribute}</code>s for
     *   this <code>Element</code>, in no
     *   particular order.
     * </p>
     *
     * @return <code>List</code> - attributes for the element.
     */
    public List getAttributes() {
        return attributes;
    }

    /**
     * <p>
     * This will return the <code>{@link Attribute}</code> for the
     *   specified attribute name on this <code>Element</code>.
     * </p>
     * <p>
     * <em>Note:</em>Currently, this does not assume that you want the
     *   default namespace.  I can perform 
     *   <code>getAttribute("foo:bar")</code>, and it does the right
     *   thing.  But if I have a document with an element 
     *   with attributes <code>foo:bar</code> as well as 
     *   <code>bar</code>, this currently gets the 
     *   <i>first</i> of those when <code>getAttribute("bar")</code> 
     *   is invoked.  Do we want to explicitly enforce the 
     *   default namespace (none) if no prefix is given? (brett)     
     * </p>     
     *
     * @param name <code>String</code> name of attribute to return.
     * @return <code>Attribute</code> - the requested attribute.
     * @throws <code>NoSuchAttributeException</code> - when the requested
     *         XML attribute does not exist for this <code>Element</code>.
     */    
    public Attribute getAttribute(String name) 
        throws NoSuchAttributeException {
        
        // If the name is a fullName (with prefix, handle it
        boolean hasNamespace = (name.indexOf(":") > 0);
        
        Iterator i = attributes.iterator();
        while (i.hasNext()) {
            Attribute att = (Attribute)i.next();
            if (hasNamespace) {
                if (att.getFullName().equals(name)) {
                    return att;
                }
            } else {
                if (att.getName().equals(name)) {
                    return att;
                }
            }
        }
        
        // If we got here, nothing found
        throw new NoSuchAttributeException(name);
    }

    /**
     * <p>
     * This will return the <code>{@link Attribute}</code> for the
     *   specified attribute name on this <code>Element</code>.
     * </p>
     *
     * @param prefix <code>String</code> prefix of attribute to return
     * @param name <code>String</code> local name of attribute to return.
     * @return <code>Attribute</code> - the requested attribute.
     * @throws <code>NoSuchAttributeException</code> - when the requested
     *         XML attribute does not exist for this <code>Element</code>.
     */    
    public Attribute getAttribute(String prefix, String name) 
        throws NoSuchAttributeException {
            
        Iterator i = attributes.iterator();
        while (i.hasNext()) {
            Attribute att = (Attribute)i.next();
            if ((att.getNamespacePrefix().equals(prefix)) &&
                (att.getName().equals(name))) {
                return att;
            }
        }
        
        // If we got here, nothing found
        throw new NoSuchAttributeException(name);
    }
    
    /**
     * <p>
     * This will set all the attributes for this <code>Element</code>.
     * </p>
     *
     * @param attributes <code>List</code> of attributes to add.
     * @return <code>Element</code> - this element modified.  
     */
    public Element setAttributes(List attributes) {
        this.attributes = attributes;
        
        return this;
    }

    /**
     * <p>
     * This will add an <code>{@link Attribute}</code> to this 
     *   <code>Element</code>.
     * </p>
     *
     * @param attribute <code>Attribute</code> to add.
     * @return <code>Element</code> - this element modified.
     */    
    public Element addAttribute(Attribute attribute) {
        attributes.add(attribute);
        
        return this;
    }
    
    /**
     * <p>
     * This will add an <code>{@link Attribute}</code> to this 
     *   <code>Element</code>.
     * </p>
     *
     * @param name name of the attribute to add.
     * @param value value of the attribute to add.
     * @return <code>Element</code> - this element modified.
     */    
    public Element addAttribute(String name, String value) {
        attributes.add(new Attribute(name, value));
        
        return this;
    }
    
    /**
     * <p>
     * This will remove the <code>Attribute</code> with the given name.
     * </p>
     * <p>
     * <em>Note:</em>Currently, this does not assume that you want the
     *   default namespace.  I can perform 
     *   <code>removeAttribute("foo:bar")</code>, and it does the right
     *   thing.  But if I have a document with an element
     *   with attributes <code>foo:bar</code> as well as 
     *   <code>bar</code>, this currently removes the 
     *   <i>first</i> of those when <code>removeAttribute("bar")</code>
     *   is invoked.  Do we want to explicitly enforce the default 
     *   namespace (none) if no prefix is given? (brett)     
     * </p>
     *
     * @param fullName <code>String</code> name of attribute to remove.
     */
    public void removeAttribute(String name) {
        // If the name is a fullName (with prefix, handle it
        boolean hasNamespace = (name.indexOf(":") > 0);
        
        Iterator i = attributes.iterator();
        while (i.hasNext()) {
            Attribute att = (Attribute)i.next();
            if (hasNamespace) {
                if (att.getFullName().equals(name)) {
                    i.remove();
                    return;
                }
            } else {
                if (att.getName().equals(name)) {
                    i.remove();
                    return;
                }
            }
        }
    }
    
    /**
     * <p>
     * This will remove the <code>Attribute</code> with the given name.
     * </p>
     *
     * @param namespacePrefix <code>String</code> prefix of
     *                  attribute to delete.
     * @param localName <code>String</code> name of attribute to remove.
     */
    public void removeAttribute(String namespacePrefix, String localName) {
        Iterator i = attributes.iterator();
        while (i.hasNext()) {
            Attribute att = (Attribute)i.next();
            if ((att.getNamespacePrefix().equals(prefix)) &&
                (att.getName().equals(name))) {
                i.remove();
                return;
            }
        }
    }
        
}